import os
import tempfile
import unittest.mock
import warnings

import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt
import numpy as np
import pytest
import torch

from cka.viz import (
    plot_cka_comparison,
    plot_cka_heatmap,
    plot_cka_layer_trend,
    plot_cka_trend,
    plot_cka_trend_with_range,
    save_figure,
)


class TestPlotCkaHeatmap:
    def test_basic_with_torch_tensor(self):
        matrix = torch.rand(5, 4)
        fig, ax = plot_cka_heatmap(matrix)

        assert fig is not None
        assert ax is not None
        plt.close(fig)

    def test_basic_with_numpy_array(self):
        matrix = np.random.rand(5, 4)
        fig, ax = plot_cka_heatmap(matrix)

        assert fig is not None
        assert ax is not None
        plt.close(fig)

    def test_custom_axis(self):
        matrix = torch.rand(5, 4)
        external_fig, external_ax = plt.subplots()

        fig, ax = plot_cka_heatmap(matrix, ax=external_ax)

        assert ax is external_ax
        assert fig is external_fig
        plt.close(fig)

    def test_annot_true(self):
        matrix = torch.rand(3, 3)

        fig, ax = plot_cka_heatmap(matrix, annot=True)

        assert len(ax.texts) == matrix.numel()
        plt.close(fig)

    def test_custom_layer_names(self):
        matrix = torch.rand(3, 2)
        layers1 = ["layer1", "layer2", "layer3"]
        layers2 = ["fc1", "fc2"]

        fig, ax = plot_cka_heatmap(matrix, layers1=layers1, layers2=layers2)

        yticklabels = [t.get_text() for t in ax.get_yticklabels()]
        xticklabels = [t.get_text() for t in ax.get_xticklabels()]
        assert yticklabels == layers1
        assert xticklabels == layers2
        plt.close(fig)

    def test_layer_name_depth(self):
        matrix = torch.rand(2, 2)
        layers1 = ["encoder.block.layer1", "encoder.block.layer2"]
        layers2 = ["decoder.fc1", "decoder.fc2"]

        fig, ax = plot_cka_heatmap(
            matrix, layers1=layers1, layers2=layers2, layer_name_depth=1
        )

        yticklabels = [t.get_text() for t in ax.get_yticklabels()]
        assert yticklabels == ["layer1", "layer2"]
        plt.close(fig)

    def test_layer_name_depth_2(self):
        matrix = torch.rand(2, 2)
        layers1 = ["encoder.block.layer1", "encoder.block.layer2"]
        layers2 = ["decoder.fc1", "decoder.fc2"]

        fig, ax = plot_cka_heatmap(
            matrix, layers1=layers1, layers2=layers2, layer_name_depth=2
        )

        yticklabels = [t.get_text() for t in ax.get_yticklabels()]
        assert yticklabels == ["block.layer1", "block.layer2"]
        plt.close(fig)

    def test_custom_vmin_vmax(self):
        matrix = torch.rand(4, 4)

        fig, ax = plot_cka_heatmap(matrix, vmin=0.2, vmax=0.8)

        images = ax.get_images()
        assert len(images) == 1
        assert images[0].get_clim() == (0.2, 0.8)
        plt.close(fig)

    def test_colorbar_false(self):
        matrix = torch.rand(4, 4)

        fig, ax = plot_cka_heatmap(matrix, colorbar=False)

        assert fig is not None
        plt.close(fig)

    def test_show_false_default(self):
        matrix = torch.rand(4, 4)

        fig, ax = plot_cka_heatmap(matrix, show=False)

        assert fig is not None
        plt.close(fig)

    def test_show_true(self):
        matrix = torch.rand(4, 4)

        with unittest.mock.patch.object(plt, "show"):
            fig, ax = plot_cka_heatmap(matrix, show=True)

        assert fig is not None
        plt.close(fig)

    def test_custom_model_names(self):
        matrix = torch.rand(4, 4)

        fig, ax = plot_cka_heatmap(
            matrix, model1_name="ResNet", model2_name="VGG", title=None
        )

        assert "ResNet vs VGG" in ax.get_title()
        plt.close(fig)

    def test_custom_title(self):
        matrix = torch.rand(4, 4)

        fig, ax = plot_cka_heatmap(matrix, title="My Custom Title")

        assert ax.get_title() == "My Custom Title"
        plt.close(fig)

    def test_custom_figsize(self):
        matrix = torch.rand(4, 4)

        fig, ax = plot_cka_heatmap(matrix, figsize=(12, 10))

        assert fig.get_size_inches()[0] == 12
        assert fig.get_size_inches()[1] == 10
        plt.close(fig)

    def test_custom_cmap(self):
        matrix = torch.rand(4, 4)

        fig, ax = plot_cka_heatmap(matrix, cmap="viridis")

        images = ax.get_images()
        assert images[0].get_cmap().name == "viridis"
        plt.close(fig)

    def test_nan_values_in_annot(self):
        matrix = torch.rand(3, 3)
        matrix[0, 0] = float("nan")

        fig, ax = plot_cka_heatmap(matrix, annot=True)

        assert len(ax.texts) == matrix.numel() - 1
        plt.close(fig)


class TestPlotCkaTrend:
    def test_single_1d_tensor(self):
        values = torch.rand(10)

        fig, ax = plot_cka_trend(values)

        assert fig is not None
        assert ax is not None
        lines = ax.get_lines()
        assert len(lines) == 1
        plt.close(fig)

    def test_2d_tensor_multiple_lines(self):
        values = torch.rand(3, 10)

        fig, ax = plot_cka_trend(values)

        lines = ax.get_lines()
        assert len(lines) == 3
        plt.close(fig)

    def test_list_of_tensors(self):
        values = [torch.rand(10), torch.rand(10)]

        fig, ax = plot_cka_trend(values)

        lines = ax.get_lines()
        assert len(lines) == 2
        plt.close(fig)

    def test_list_of_numpy_arrays(self):
        values = [np.random.rand(10), np.random.rand(10)]

        fig, ax = plot_cka_trend(values)

        lines = ax.get_lines()
        assert len(lines) == 2
        plt.close(fig)

    def test_custom_x_values(self):
        values = torch.rand(5)
        x_values = [0, 2, 4, 6, 8]

        fig, ax = plot_cka_trend(values, x_values=x_values)

        lines = ax.get_lines()
        xdata = lines[0].get_xdata()
        assert list(xdata) == x_values
        plt.close(fig)

    def test_custom_colors(self):
        values = [torch.rand(10), torch.rand(10)]
        colors = ["red", "blue"]

        fig, ax = plot_cka_trend(values, colors=colors)

        assert fig is not None
        plt.close(fig)

    def test_custom_linestyles(self):
        values = [torch.rand(10), torch.rand(10)]
        linestyles = ["--", "-."]

        fig, ax = plot_cka_trend(values, linestyles=linestyles)

        assert fig is not None
        plt.close(fig)

    def test_custom_markers(self):
        values = [torch.rand(10), torch.rand(10)]
        markers = ["s", "^"]

        fig, ax = plot_cka_trend(values, markers=markers)

        assert fig is not None
        plt.close(fig)

    def test_labels(self):
        values = [torch.rand(10), torch.rand(10)]
        labels = ["Model A", "Model B"]

        fig, ax = plot_cka_trend(values, labels=labels, legend=True)

        legend = ax.get_legend()
        assert legend is not None
        plt.close(fig)

    def test_legend_true_multiple_lines(self):
        values = [torch.rand(10), torch.rand(10)]

        fig, ax = plot_cka_trend(values, legend=True)

        legend = ax.get_legend()
        assert legend is not None
        plt.close(fig)

    def test_legend_false(self):
        values = [torch.rand(10), torch.rand(10)]

        fig, ax = plot_cka_trend(values, legend=False)

        legend = ax.get_legend()
        assert legend is None
        plt.close(fig)

    def test_legend_single_line(self):
        values = torch.rand(10)

        fig, ax = plot_cka_trend(values, legend=True)

        legend = ax.get_legend()
        assert legend is None
        plt.close(fig)

    def test_grid_false(self):
        values = torch.rand(10)

        fig, ax = plot_cka_trend(values, grid=False)

        assert fig is not None
        plt.close(fig)

    def test_custom_ax(self):
        values = torch.rand(10)
        external_fig, external_ax = plt.subplots()

        fig, ax = plot_cka_trend(values, ax=external_ax)

        assert ax is external_ax
        assert fig is external_fig
        plt.close(fig)

    def test_custom_xlabel_ylabel(self):
        values = torch.rand(10)

        fig, ax = plot_cka_trend(values, xlabel="Epoch", ylabel="Similarity")

        assert ax.get_xlabel() == "Epoch"
        assert ax.get_ylabel() == "Similarity"
        plt.close(fig)

    def test_custom_title(self):
        values = torch.rand(10)

        fig, ax = plot_cka_trend(values, title="Trend Plot")

        assert ax.get_title() == "Trend Plot"
        plt.close(fig)

    def test_ylim_set(self):
        values = torch.rand(10)

        fig, ax = plot_cka_trend(values)

        ylim = ax.get_ylim()
        assert ylim[0] == 0
        assert ylim[1] == 1.05
        plt.close(fig)

    def test_1d_numpy_array(self):
        values = np.random.rand(10)

        fig, ax = plot_cka_trend(values)

        lines = ax.get_lines()
        assert len(lines) == 1
        plt.close(fig)

    def test_show_true(self):
        values = torch.rand(10)

        with unittest.mock.patch.object(plt, "show"):
            fig, ax = plot_cka_trend(values, show=True)

        assert fig is not None
        plt.close(fig)


class TestPlotCkaTrendWithRange:
    def test_single_group_2d(self):
        values = torch.rand(4, 10)

        fig, ax = plot_cka_trend_with_range(values)

        lines = ax.get_lines()
        assert len(lines) == 1
        assert len(ax.collections) > 0
        plt.close(fig)

    def test_multiple_groups_3d(self):
        values = torch.rand(3, 4, 10)

        fig, ax = plot_cka_trend_with_range(values)

        lines = ax.get_lines()
        assert len(lines) == 3
        assert len(ax.collections) >= 3
        plt.close(fig)


class TestPlotCkaLayerTrend:
    def test_single_matrix(self):
        matrix = torch.rand(5, 5)
        layers = [f"layer{i}" for i in range(5)]

        fig, ax = plot_cka_layer_trend(matrix, layers=layers)

        assert fig is not None
        xticklabels = [t.get_text() for t in ax.get_xticklabels()]
        assert xticklabels == layers
        plt.close(fig)

class TestPlotCkaComparison:
    def test_basic_comparison(self):
        matrices = [torch.rand(4, 4), torch.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles)

        assert fig is not None
        assert axes.shape == (1, 2)
        plt.close(fig)

    def test_share_colorbar_true(self):
        matrices = [torch.rand(4, 4), torch.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles, share_colorbar=True)

        assert fig is not None
        plt.close(fig)

    def test_share_colorbar_false(self):
        matrices = [torch.rand(4, 4), torch.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles, share_colorbar=False)

        assert fig is not None
        plt.close(fig)

    def test_custom_ncols(self):
        matrices = [torch.rand(4, 4) for _ in range(4)]
        titles = [f"Matrix {i}" for i in range(4)]

        fig, axes = plot_cka_comparison(matrices, titles, ncols=4)

        assert axes.shape == (1, 4)
        plt.close(fig)

    def test_auto_figsize(self):
        matrices = [torch.rand(4, 4), torch.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles, figsize=None)

        assert fig is not None
        plt.close(fig)

    def test_custom_figsize(self):
        matrices = [torch.rand(4, 4), torch.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles, figsize=(15, 8))

        size = fig.get_size_inches()
        assert size[0] == 15
        assert size[1] == 8
        plt.close(fig)

    def test_hide_empty_subplots(self):
        matrices = [torch.rand(4, 4) for _ in range(3)]
        titles = [f"Matrix {i}" for i in range(3)]

        fig, axes = plot_cka_comparison(matrices, titles, ncols=2)

        assert axes.shape == (2, 2)
        assert not axes[1, 1].get_visible()
        plt.close(fig)

    def test_heatmap_kwargs_passthrough(self):
        matrices = [torch.rand(4, 4), torch.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles, annot=True, tick_fontsize=10)

        assert fig is not None
        plt.close(fig)

    def test_custom_layers(self):
        matrices = [torch.rand(3, 3), torch.rand(3, 3)]
        titles = ["Matrix 1", "Matrix 2"]
        layers = ["layer1", "layer2", "layer3"]

        fig, axes = plot_cka_comparison(matrices, titles, layers=layers)

        assert fig is not None
        plt.close(fig)

    def test_numpy_arrays(self):
        matrices = [np.random.rand(4, 4), np.random.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles)

        assert fig is not None
        plt.close(fig)

    def test_mixed_torch_numpy(self):
        matrices = [torch.rand(4, 4), np.random.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        fig, axes = plot_cka_comparison(matrices, titles)

        assert fig is not None
        plt.close(fig)

    def test_single_matrix(self):
        matrices = [torch.rand(4, 4)]
        titles = ["Matrix 1"]

        fig, axes = plot_cka_comparison(matrices, titles)

        assert fig is not None
        plt.close(fig)

    def test_show_true(self):
        matrices = [torch.rand(4, 4), torch.rand(4, 4)]
        titles = ["Matrix 1", "Matrix 2"]

        with unittest.mock.patch.object(plt, "show"):
            fig, axes = plot_cka_comparison(matrices, titles, show=True)

        assert fig is not None
        plt.close(fig)


class TestSaveFigure:
    def test_basic_save(self):
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3], [1, 2, 3])

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "test.png")
            save_figure(fig, path)

            assert os.path.exists(path)
            assert os.path.getsize(path) > 0

    def test_custom_dpi(self):
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3], [1, 2, 3])

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "test_dpi.png")
            save_figure(fig, path, dpi=300)

            assert os.path.exists(path)

    def test_custom_bbox_inches(self):
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3], [1, 2, 3])

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "test_bbox.png")
            save_figure(fig, path, bbox_inches="tight")

            assert os.path.exists(path)

    def test_transparent(self):
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3], [1, 2, 3])

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "test_transparent.png")
            save_figure(fig, path, transparent=True)

            assert os.path.exists(path)

    def test_figure_closed_after_save(self):
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3], [1, 2, 3])
        fig_num = fig.number

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "test_close.png")
            save_figure(fig, path)

            assert fig_num not in plt.get_fignums()

    def test_save_pdf(self):
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3], [1, 2, 3])

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "test.pdf")
            save_figure(fig, path)

            assert os.path.exists(path)



import pytest
import torch

from cka.hsic import hsic, hsic_cross


class TestHsic:
    def test_valid_3d_inputs(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 8, 8)

        hsic_xy, hsic_xx, hsic_yy = hsic(grams_x, grams_y)

        assert hsic_xy.shape == (5, 3)
        assert hsic_xx.shape == (5,)
        assert hsic_yy.shape == (3,)

    def test_non_3d_input_x_raises_error(self):
        grams_x = torch.randn(5, 8)
        grams_y = torch.randn(3, 8, 8)

        with pytest.raises(ValueError, match="hsic_all requires 3D tensors"):
            hsic(grams_x, grams_y)

    def test_non_3d_input_y_raises_error(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 8, 8, 2)

        with pytest.raises(ValueError, match="hsic_all requires 3D tensors"):
            hsic(grams_x, grams_y)

    def test_non_square_matrix_x_raises_error(self):
        grams_x = torch.randn(5, 8, 6)
        grams_y = torch.randn(3, 8, 8)

        with pytest.raises(ValueError, match="hsic_all requires square matrices"):
            hsic(grams_x, grams_y)

    def test_non_square_matrix_y_raises_error(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 8, 6)

        with pytest.raises(ValueError, match="hsic_all requires square matrices"):
            hsic(grams_x, grams_y)

    def test_mismatched_sample_dimension_raises_error(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 10, 10)

        with pytest.raises(ValueError, match="must have same sample dimension"):
            hsic(grams_x, grams_y)

    def test_n_equals_3_raises_error(self):
        grams_x = torch.randn(5, 3, 3)
        grams_y = torch.randn(3, 3, 3)

        with pytest.raises(ValueError, match="hsic_all requires n > 3"):
            hsic(grams_x, grams_y)

    def test_n_equals_2_raises_error(self):
        grams_x = torch.randn(5, 2, 2)
        grams_y = torch.randn(3, 2, 2)

        with pytest.raises(ValueError, match="hsic_all requires n > 3"):
            hsic(grams_x, grams_y)

    def test_diagonal_is_one_for_normalized_gram(self):
        n_samples = 8
        n_layers = 4

        features = torch.randn(n_layers, n_samples, 16)
        features = features / features.norm(dim=-1, keepdim=True)
        grams = torch.bmm(features, features.transpose(1, 2))

        hsic_xy, hsic_xx, hsic_yy = hsic(grams, grams)

        diagonal = torch.diagonal(hsic_xy)
        assert torch.allclose(diagonal, hsic_xx, atol=1e-5)
        assert torch.allclose(diagonal, hsic_yy, atol=1e-5)


class TestHsicOuter:
    def test_valid_inputs_shape(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 8, 8)

        result = hsic_cross(grams_x, grams_y)

        assert result.shape == (5, 3)

    def test_non_3d_input_x_raises_error(self):
        grams_x = torch.randn(5, 8)
        grams_y = torch.randn(3, 8, 8)

        with pytest.raises(ValueError, match="hsic_cross requires 3D tensors"):
            hsic_cross(grams_x, grams_y)

    def test_non_3d_input_y_raises_error(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 8, 8, 2)

        with pytest.raises(ValueError, match="hsic_cross requires 3D tensors"):
            hsic_cross(grams_x, grams_y)

    def test_non_square_matrix_x_raises_error(self):
        grams_x = torch.randn(5, 8, 6)
        grams_y = torch.randn(3, 8, 8)

        with pytest.raises(ValueError, match="hsic_cross requires square matrices"):
            hsic_cross(grams_x, grams_y)

    def test_non_square_matrix_y_raises_error(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 8, 6)

        with pytest.raises(ValueError, match="hsic_cross requires square matrices"):
            hsic_cross(grams_x, grams_y)

    def test_mismatched_sample_dimension_raises_error(self):
        grams_x = torch.randn(5, 8, 8)
        grams_y = torch.randn(3, 10, 10)

        with pytest.raises(ValueError, match="must have same sample dimension"):
            hsic_cross(grams_x, grams_y)

    def test_n_equals_3_raises_error(self):
        grams_x = torch.randn(5, 3, 3)
        grams_y = torch.randn(3, 3, 3)

        with pytest.raises(ValueError, match="hsic_cross requires n > 3"):
            hsic_cross(grams_x, grams_y)

    def test_n_equals_1_raises_error(self):
        grams_x = torch.randn(5, 1, 1)
        grams_y = torch.randn(3, 1, 1)

        with pytest.raises(ValueError, match="hsic_cross requires n > 3"):
            hsic_cross(grams_x, grams_y)

    def test_symmetry_same_inputs(self):
        grams = torch.randn(4, 8, 8)
        grams = grams + grams.transpose(-1, -2)

        result = hsic_cross(grams, grams)

        assert torch.allclose(result, result.T, atol=1e-5)

    def test_consistency_with_hsic(self):
        grams_x = torch.randn(3, 8, 8)
        grams_y = torch.randn(4, 8, 8)

        grams_x = grams_x + grams_x.transpose(-1, -2)
        grams_y = grams_y + grams_y.transpose(-1, -2)

        outer_result = hsic_cross(grams_x, grams_y)
        hsic_xy, _, _ = hsic(grams_x, grams_y)

        assert torch.allclose(outer_result, hsic_xy, atol=1e-5)



import torch


def hsic(
    grams_x: torch.Tensor,
    grams_y: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    if grams_x.dim() != 3 or grams_y.dim() != 3:
        raise ValueError(
            f"hsic_all requires 3D tensors, got shapes {grams_x.shape} and {grams_y.shape}"
        )

    _, n, m = grams_x.shape
    _, n_y, m_y = grams_y.shape

    if n != m or n_y != m_y:
        raise ValueError("hsic_all requires square matrices")

    if n != n_y:
        raise ValueError(
            f"Gram matrices must have same sample dimension, got {n} and {n_y}"
        )

    if n <= 3:
        raise ValueError(f"hsic_all requires n > 3, got n={n}")

    # Shared intermediates for grams_x
    diag_x = torch.diagonal(grams_x, dim1=-2, dim2=-1)  # (n1, n)
    sum_x = grams_x.sum(dim=(-2, -1)) - diag_x.sum(dim=-1)  # (n1,)
    col_sum_x = grams_x.sum(dim=-2) - diag_x  # (n1, n)

    # Shared intermediates for grams_y
    diag_y = torch.diagonal(grams_y, dim1=-2, dim2=-1)  # (n2, n)
    sum_y = grams_y.sum(dim=(-2, -1)) - diag_y.sum(dim=-1)  # (n2,)
    col_sum_y = grams_y.sum(dim=-2) - diag_y  # (n2, n)

    denominator = n * (n - 3)
    term2_denom = (n - 1) * (n - 2)
    term3_coeff = 2 / (n - 2)

    # Cross HSIC (hsic_xy)
    trace_xy = torch.einsum("aij,bij->ab", grams_x, grams_y) - torch.einsum(
        "ai,bi->ab", diag_x, diag_y
    )
    term2_xy = torch.outer(sum_x, sum_y) / term2_denom
    term3_xy = term3_coeff * torch.mm(col_sum_x, col_sum_y.T)
    hsic_xy = (trace_xy + term2_xy - term3_xy) / denominator

    # Self HSIC for grams_x (hsic_xx)
    trace_xx = (grams_x * grams_x).sum(dim=(-2, -1)) - (diag_x * diag_x).sum(dim=-1)
    term2_xx = (sum_x * sum_x) / term2_denom
    term3_xx = term3_coeff * (col_sum_x * col_sum_x).sum(dim=-1)
    hsic_xx = (trace_xx + term2_xx - term3_xx) / denominator

    # Self HSIC for grams_y (hsic_yy)
    trace_yy = (grams_y * grams_y).sum(dim=(-2, -1)) - (diag_y * diag_y).sum(dim=-1)
    term2_yy = (sum_y * sum_y) / term2_denom
    term3_yy = term3_coeff * (col_sum_y * col_sum_y).sum(dim=-1)
    hsic_yy = (trace_yy + term2_yy - term3_yy) / denominator

    return hsic_xy, hsic_xx, hsic_yy


def hsic_cross(
    grams_x: torch.Tensor,
    grams_y: torch.Tensor,
) -> torch.Tensor:
    if grams_x.dim() != 3 or grams_y.dim() != 3:
        raise ValueError(
            f"hsic_cross requires 3D tensors, got shapes {grams_x.shape} and {grams_y.shape}"
        )

    _, n, m = grams_x.shape
    _, n_y, m_y = grams_y.shape

    if n != m or n_y != m_y:
        raise ValueError("hsic_cross requires square matrices")

    if n != n_y:
        raise ValueError(
            f"Gram matrices must have same sample dimension, got {n} and {n_y}"
        )

    if n <= 3:
        raise ValueError(f"hsic_cross requires n > 3, got n={n}")

    diag_x = torch.diagonal(grams_x, dim1=-2, dim2=-1)  # (n1, n)
    diag_y = torch.diagonal(grams_y, dim1=-2, dim2=-1)  # (n2, n)

    # Term 1: trace_KL[i,j] = sum(grams_x[i] * grams_y[j]) - sum(diag_x[i] * diag_y[j])
    # Use einsum for pairwise element-wise product sum
    trace_full = torch.einsum("aij,bij->ab", grams_x, grams_y)  # (n1, n2)
    trace_diag = torch.einsum("ai,bi->ab", diag_x, diag_y)  # (n1, n2)
    trace_KL = trace_full - trace_diag

    # Term 2: (sum_K[i] * sum_L[j]) / ((n-1)(n-2))
    sum_K = grams_x.sum(dim=(-2, -1)) - diag_x.sum(dim=-1)  # (n1,)
    sum_L = grams_y.sum(dim=(-2, -1)) - diag_y.sum(dim=-1)  # (n2,)
    term2 = torch.outer(sum_K, sum_L) / ((n - 1) * (n - 2))  # (n1, n2)

    # Term 3: 2 * (col_sum_K[i] @ col_sum_L[j]) / (n-2)
    col_sum_K = grams_x.sum(dim=-2) - diag_x  # (n1, n)
    col_sum_L = grams_y.sum(dim=-2) - diag_y  # (n2, n)
    term3 = 2 * torch.mm(col_sum_K, col_sum_L.T) / (n - 2)  # (n1, n2)

    main_term = trace_KL + term2 - term3
    denominator = n * (n - 3)
    return main_term / denominator
